<?php declare(strict_types = 1);

namespace Drupal\disable_libraries\Form;

use Drupal\Core\Form\ConfigFormBase;
use Drupal\Core\Form\FormStateInterface;

/**
 * Configure disable_libraries settings for this site.
 */
final class SettingsForm extends ConfigFormBase {

  /**
   * List mode select box options
   */
  public $modeOptions = [
    'allow' => 'allow',
    'deny'  => 'deny',
  ];

  /**
   * Convert newline separated string to array
   */
  public function splitByNewLine(string $text): array {
    return preg_split('~\R+~u', $text, -1, PREG_SPLIT_NO_EMPTY);
  }

  /**
   * {@inheritdoc}
   */
  public function getFormId(): string {
    return 'disable_libraries_settings';
  }

  /**
   * {@inheritdoc}
   */
  protected function getEditableConfigNames(): array {
    return ['disable_libraries.settings'];
  }

  /**
   * {@inheritdoc}
   */
  public function buildForm(array $form, FormStateInterface $form_state): array {

    // form api: https://www.drupal.org/docs/drupal-apis/form-api/form-render-elements
    // form types: https://api.drupal.org/api/drupal/namespace/Drupal%21Core%21Render%21Element/8.2.x

    $config = $this->config('disable_libraries.settings');

    $form['anonymous'] = [
      '#type' => 'details',
      '#title' => t('Config for anonymous users'),
      '#open' => TRUE,
    ];
    $form['anonymous']['anonMode'] = [
      '#type' => 'select',
      '#options' => $this->modeOptions,
      '#title' => $this->t('List mode for anonymous users'),
      '#description' => $this->t('With an allow list all libraries, except allowed ones, are disabled. With a deny list, all libraries, except denied ones, are enabled.<br />Default: allow'),
      '#default_value' => $config->get('anonMode'),
    ];
    $value = $config->get('deniedAnon') ?? [];
    $value = implode("\r\n", $value);
    $form['anonymous']['deniedAnon'] = [
      '#type' => 'textarea',
      '#title' => $this->t('Deny list for anonymous users'),
      '#description' => $this->t('List of denied modules. Write each module on a separate line.'),
      '#default_value' => $value,
    ];
    $value = $config->get('allowedAnon') ?? [];
    $value = implode("\r\n", $value);
    $form['anonymous']['allowedAnon'] = [
      '#type' => 'textarea',
      '#title' => $this->t('Allow list for anonymous users'),
      '#description' => $this->t('List of allowed modules. Write each module on a separate line.'),
      '#default_value' => $value,
    ];

    $form['authenticated'] = [
      '#type' => 'details',
      '#title' => t('Config for authenticated users'),
      '#open' => TRUE,
    ];
    $form['authenticated']['authMode'] = [
      '#type' => 'select',
      '#options' => $this->modeOptions,
      '#title' => $this->t('List mode for authenticated users'),
      '#description' => $this->t('With an allow list all libraries, except allowed ones, are disabled. With a deny list all libraries, except denied ones, are enabled.<br />Default: deny'),
      '#default_value' => $config->get('authMode'),
    ];
    $value = $config->get('deniedAuth') ?? [];
    $value = implode("\r\n", $value);
    $form['authenticated']['deniedAuth'] = [
      '#type' => 'textarea',
      '#title' => $this->t('Deny list for authenticated users'),
      '#description' => $this->t('List of denied modules. Write each module on a separate line.'),
      '#default_value' => $value,
    ];
    $value = $config->get('allowedAuth') ?? [];
    $value = implode("\r\n", $value);
    $form['authenticated']['allowedAuth'] = [
      '#type' => 'textarea',
      '#title' => $this->t('Allow list for authenticated users'),
      '#description' => $this->t('List of allowed modules. Write each module on a separate line.'),
      '#default_value' => $value,
    ];

    $form['other'] = [
      '#type' => 'details',
      '#title' => t('Other config'),
      '#open' => TRUE,
    ];
    $form['other']['allowThemeAutomatically'] = [
      '#type' => 'checkbox',
      '#title' => $this->t('Automatically allow libraries from active theme and parent themes.'),
      '#description' => $this->t('Disabling this setting is a bad idea and things might break.<br />Default: enabled.'),
      '#default_value' => $config->get('allowThemeAutomatically'),
    ];
    $form['other']['disableCDN'] = [
      '#type' => 'checkbox',
      '#title' => $this->t('Disable CDN keys of all modules.'),
      '#description' => $this->t('Some modules (e. g. webform) have CDNs enabled by default. To prevent GDPR or security disasters, they should be disabled.<br />Default: enabled.'),
      '#default_value' => $config->get('disableCDN'),
    ];

    // TODO: not needed anymore, because of drupal_flush_all_caches() after form submit
    // $url = \Drupal\Core\Url::fromRoute('system.performance_settings')->toString();
    // $form_state->getFormObject()->messenger()->addWarning($this->t('Clear the cache after any change on the <a href="@url">Performance settings page</a> to take effect or run <code>drush cache:rebuild</code> on the command line.', ['@url' => $url]));

    return parent::buildForm($form, $form_state);
  }

  /**
   * {@inheritdoc}
   */
  public function validateForm(array &$form, FormStateInterface $form_state): void {
    // @todo Validate the form here.
    // Example:
    // @code
    //   if ($form_state->getValue('example') === 'wrong') {
    //     $form_state->setErrorByName(
    //       'message',
    //       $this->t('The value is not correct.'),
    //     );
    //   }
    // @endcode
    parent::validateForm($form, $form_state);
  }

  /**
   * {@inheritdoc}
   */
  public function submitForm(array &$form, FormStateInterface $form_state): void {

    $config = $this->config('disable_libraries.settings');

    $hasChanged = false;

    $configNames = [
      'authMode',
      'anonMode',
    ];
    foreach ($configNames as $name) {
      $value = $form_state->getValue($name);
      if ($config->get($name) !== $value) $hasChanged = true;
      $config->set($name, $value);
    }

    $configNames = [
      'allowThemeAutomatically',
      'disableCDN',
    ];
    foreach ($configNames as $name) {
      $value = (bool) $form_state->getValue($name);
      if ($config->get($name) !== $value) $hasChanged = true;
      $config->set($name, $value);
    }

    $configNames = [
      'allowedAuth',
      'allowedAnon',
      'deniedAuth',
      'deniedAnon',
    ];
    foreach ($configNames as $name) {
      $value = $this->splitByNewLine($form_state->getValue($name));
      if ($config->get($name) !== $value) $hasChanged = true;
      $config->set($name, $value);
    }

    $config->save();

    // flush cache automatically
    if ($hasChanged) \drupal_flush_all_caches();

    parent::submitForm($form, $form_state);
  }

}
